--[==[
--------------------------------------------------------------------------------
--* Name:        startup.lua
--* Version:     v3.0
--* Purpose:     Script manager & utilities for Direct Connect clients
--* Author:      sedulus (I guess)
--* Modified by: Thor [ejjeliorjarat@gmail.com]
--* Created:     2010.03.02
--* Copyright:   (c) Thor 2010
                 Based on RSX++ new (but removed) Lua API by adrian_007 (c) 2009
				 Rewritten ScriptManager part by ATAG (c) 2008
				 Some dcpp.Utils function from DC++ by arnetheduck (c) 2001-2010
				 alien v0.5.0.1 (c) Fabio Mascarenhas
--* Licence:     GNU-GPLv2
--------------------------------------------------------------------------------
--* Changelog:
--* Added alien library
--* OO call implementation & instances
--* Function argument checks
--* dco object instead of calling DC()
--* New function names
--* Useful new calls and the old ones moved to dcpp.xxx tables
--* Implemented privacy: can't use/modify private fields (in hub,user,dcpp tables)
--* Using metatable on string functions
--* Simplified functions with booleans
--* New calls in hub and user tables
--* Locale
--* UserCommand (on ADC hubs too)
--* Lua manager: easy load/unload/reload/list of script, list memory usage
--* Scripts can use global variables, it won't cause name clash (diff. envs)
--* Supporting locale characters in filenames
--* Full html documentation
--------------------------------------------------------------------------------
]==]

-- forward compatibility (Lua5.2)
local unpack = unpack or table.unpack
local debug = require("debug")

local dco = DC() -- Create an instance instead of call the function every time
DC = nil

print = function(msg,...)
	msg = dco:ToUtf8(msg)
	dco:PrintDebug(tostring(msg):format(...))
end

--[=[----------------]=]--
--[=[ Load libraries ]=]--
--[=[----------------]=]--

package.cpath = "./scripts/lib/?.dll;"..package.cpath
package.path = "./scripts/lib/?.lua;"..package.path

require("alien")
--[=[
local libgeoip = nil

do
	require("geoip")
	libgeoip = geoip
	local ok, err = libgeoip.open(dco:GetAppPath().."scripts\\lib\\GeoIP.dat")
	if not ok then
		print("Unable to load GeoIP.dat file from the scripts\\lib directory. Please "..
		"download the geoIP.dat file from http://www.maxmind.com/app/geoip_country") 
	end
	geoip = nil
end]=]

os.setlocale("")
os.setlocale("C","numeric")

local _ENV = {}

function _ALERT(s)
	local errormsg,fn = s..debug.traceback(),nil
	for script in errormsg:gmatch("scripts\\(.+%).lua") do
		if script ~= "startup.lua" then fn = script break end -- Find where it came from
	end
	if _ENV[fn] and _ENV[fn]["OnError"] then
		_ENV[fn]["OnError"](s)
	else
		local f = io.open("scriptlog.txt","a")
		f:write(os.date("[%x %X] ")..errormsg.."\n")
		print(errormsg)
		f:close()
	end
end

local argcheck = function(fname,ipos,pname,ptype)
	table.insert(ptype,"")
	local ok = nil
	for i,v in ipairs(ptype) do
		if type(pname) == v then
			ok = true; break
		end
	end
	if not ok then
		_ALERT(("bad argument #%d to %s (%s expected, got %s)"):format(ipos,fname,table.concat(ptype," or "):sub(1,-5),type(pname)))
	end
end

--[=[---------------------]=]--
--[=[ Creating dcpp table ]=]--
--[=[---------------------]=]--

local _dcpp = {_start = os.time(),_hubs = {}, _listeners = {}, _usercommands = {}, _oplists = {}} -- private storage

local _dcppmeta = {}
if not dcpp or dcpp._init_me_anyway then
	dcpp = {} -- global tables/calls
	_dcppmeta.AdcCommand = {}
	_dcppmeta.ClientType = {
		BOT    = 0x01,
		REGGED = 0x02,
		OP     = 0x04,
		SU     = 0x08,
		OWNER  = 0x10,
		HUB    = 0x20,
	}
	_dcppmeta.FavoriteManager = {}
	_dcppmeta.ClientManager = {}
	_dcppmeta.SettingsManager = {}
	_dcppmeta.StatusFlags = {
		NORMAL   = 0x01,
		AWAY     = 0x02,
		SERVER   = 0x04,
		FIREBALL = 0x08,
		TLS      = 0x10,
		NAT      = 0x20,
	}
	_dcppmeta.Text = {}
	_dcppmeta.UCContext = {
		HUB      = 0x01,
		CHAT     = 0x02,
		SEARCH   = 0x04,
		FILELIST = 0x08,
		MASK     = 0x0F,
	}
	_dcppmeta.UCType = {
		SEPARATOR = 0x00,
		RAW       = 0x01,
		RAW_ONCE  = 0x10,
		REMOVE    = 0x11,
		CLEAR     = 0xFF,
	}
	_dcppmeta.Utils = {}	
end

_dcppmeta.LogMessage = function(msg)
	dco:PrintDebug(msg)
end

-- AdcCommand

_dcppmeta.AdcCommand.toSID = function(sid)
	argcheck("dcpp.AdcCommand.toSID",1,sid,{"string"})
	local j = 0
	for i=1,#sid do j = j + sid:sub(i,i):byte()*256^(i-1) end
	return j
end

_dcppmeta.AdcCommand.fromSID = function(sid)
	argcheck("dcpp.AdcCommand.fromSID",1,sid,{"number"})
	return string.format("%c%c%c%c",sid%0x100,sid/0x100%0x100,sid/0x10000%0x100,sid/0x1000000%0x100)
end

-- ClientManager

_dcppmeta.ClientManager.closeClient = function(hub)
	argcheck("dcpp.ClientManager.closeClient",1,hub,{"userdata"})
	return dco:DeleteClient(hub)
end

_dcppmeta.ClientManager.getHubs = function()
	local t = {}
	for a,b in pairs(_dcpp._hubs) do table.insert(t,a) end
	return t
end

_dcppmeta.ClientManager.getUsers = function()
	local t = {}
	for a,b in pairs(_dcpp._hubs) do
		for id,user in pairs(b:getUserList()) do
			table.insert(t,user)
		end
	end
	return t
end

_dcppmeta.ClientManager.findHub = function(hub)
	argcheck("dcpp.ClientManager.findHub",1,hub,{"userdata","string"})
	for k,h in pairs(_dcpp._hubs) do
		if k == hub or (h:getHubIdentity() and (h:getHubIdentity():get("URL") == hub or h:getHubUrl() == hub)) then
			return h
		end
	end
	return nil
end

_dcppmeta.ClientManager.openClient = function(huburl,nick)
	argcheck("dcpp.ClientManager.openClient",1,huburl,{"string"})
	argcheck("dcpp.ClientManager.openClient",2,nick,{"string"})
	return dco:CreateClient(huburl,nick)
end

-- FavoriteManager

local sendUserCommands = function(hub,id)
	hub = dcpp.ClientManager.findHub(hub)
	if hub:isAdc() then
		hub:parseCommand("ICMD ".._dcpp._usercommands[id][1]:getName().." RM1")
		_dcpp._usercommands[id] = nil
	else
		_dcpp._usercommands[id] = nil
		hub:parseCommand("$UserCommand 255 15") -- On NMDC, remove all, then re-add all -.-'
		for i,ucmd in ipairs(dcpp.FavoriteManager.get()) do
			local _hub = ucmd:getHub()
			if ((_hub == hub:getHubIdentity():get("URL")) or (_hub == "") or (_hub == "op" and hub:isOp())) then
				hub:parseCommand(("$UserCommand %d %d %s$%s"):format(ucmd:getType(),ucmd:getContext(),ucmd:getName(),ucmd:getCommand()))
			end
		end
	end
end

_dcppmeta.FavoriteManager.addUserCommand = function(UCType,UCContext,name,command,hub,hubsent)
	argcheck("dcpp.FavoriteManager.addUserCommand",1,UCType,{"number"})
	argcheck("dcpp.FavoriteManager.addUserCommand",2,UCContext,{"number"})
	argcheck("dcpp.FavoriteManager.addUserCommand",3,name,{"string"})
	argcheck("dcpp.FavoriteManager.addUserCommand",4,command,{"string"})
	argcheck("dcpp.FavoriteManager.addUserCommand",5,hub,{"string","table"})
	if type(hub) == "table" and hubsent then
		if hub:isAdc() then
			hub:parseCommand("ICMD "..hub:escape(name).." TT"..hub:escape(command).." CT"..UCContext)
		else
			hub:parseCommand("$UserCommand "..UCType.." "..UCContext.." "..hub:escape(name).."$"..hub:escape(command))
		end
	end
	local file = ""
	if not hubsent then -- We should modify it a bit(?)
		file = debug.traceback():match("in function 'addUserCommand'.-scripts\\(%w+)%.lua:%d+:") -- Which script sent it
		command = command:gsub("%%%[(.-)%]","!%%!{%1!}")
		command = "%[lua:_ENV."..file..".OnOwnMessage(\"\",\""..command.."\")]&#124;|"
	end
	
	local id = #_dcpp._usercommands+1
	local getId = function(self) return id end
	
	local getCommand = function(self) return command end
	
	local getContext = function(self) return UCContext end

	local getHub = function(self) return hub end
	
	local getName = function(self) return name end

	local getType = function(self) return UCType end

	local isHub = function(self) return hubsent end

	local remove = function(self) sendUserCommands(self:getHub():lower(),getId()) end
	
	local usercommandinstance = {
	getId = getId,	   getCommand = getCommand, getContext = getContext, getHub = getHub, 
	getName = getName, getType = getType,       isHub = isHub,           remove = remove,
	}
	table.insert(_dcpp._usercommands,{usercommandinstance,file})

	return usercommandinstance
end

_dcppmeta.FavoriteManager.getUserCommands = function()
	local t = {}
	for _,uc in dcpp.Utils.sortByKeys(_dcpp._usercommands) do
		table.insert(t,uc)
	end
	return t
end

-- SettingsManager

_dcppmeta.SettingsManager.get = function(str)
	argcheck("dcpp.SettingsManager.get",1,str,{"string"})
	return dco:GetSetting(str)
end

-- Text

_dcppmeta.Text.acpToUtf8 = function(text)
	argcheck("dcpp.Text.acpToUtf8",1,text,{"string"})
	return dco:ToUtf8(text)
end

_dcppmeta.Text.isAscii = function(text)
	argcheck("dcpp.Text.isAscii",1,text,{"string"})
	for i=1,#text do
		if text:sub(i,i):byte() > 0x80 then return true end
	end
	return false
end

_dcppmeta.Text.tokenize = function(str,c)
	argcheck("dcpp.Text.tokenize",1,str,{"string"})
	argcheck("dcpp.Text.tokenize",2,c,{"string","nil"})
	c = c or " "
	local t = {}
	for m in str:gmatch("([^"..c.."]+)") do table.insert(t,m) end
	return t
end

_dcppmeta.Text.utf8ToAcp = function(text)
	argcheck("dcpp.Text.utf8ToAcp",1,text,{"string"})
	return dco:FromUtf8(text)
end

-- Utils

_dcppmeta.Utils.decToIp = function(ip)
	argcheck("dcpp.Utils.decToIp",1,ip,{"number"})
	local i = ip%0x1000000;
	local j = i%0x10000;
	return ("%d.%d.%d.%d"):format(ip/0x1000000,i/0x10000,j/0x100,j%0x100);
end

_dcppmeta.Utils.encodeURI = function(str,rev)
	argcheck("dcpp.Utils.encodeURI",1,str,{"string"})
	if rev then
		str = str:gsub("%+"," ")
		str = str:gsub("%%(%x%x)",function(hex) return string.char(tonumber(hex,0x10)) end)
		return str
	else
		str = str:gsub("([^abcdefghijklmnopqrstuvwxyz0123456789_%-%. ])",function(s) return string.format("%%%02X",s:byte()) end)
		str = str:gsub(" ","+")
		return str
	end
end

_dcppmeta.Utils.fileExists = function(filepath)
	argcheck("_dcppmeta.Utils.fileExists",1,filepath,{"string"})
	local f = io.open(_dcppmeta.Utils.validateFileName(filepath))
	if f then f:close() return true end
	return false
end

_dcppmeta.Utils.forFiles = function(path,pattern)
	if not path:sub(-1,-1):match("[\\/]") then
		path = path.."/"
	end
	local finddata_t = alien.defstruct{
		{"attrib","uint"},
		{"time_create","long"},
		{"time_access","long"},
		{"time_write","long"},
		{"size","ulong"},
		{"name","char"},
	}
	finddata_t.size = finddata_t.size+260
	local _finddata_t = finddata_t:new()

	local function getfilename(fd)
		local out = {}
		local offset = finddata_t.offsets.name
		local buf = fd()
		for i = offset+1, offset+260 do
			local c = buf:get(i, "char")
			if c > 0 then out[#out+1] = string.char(c)
			else break end
		end
		return table.concat(out)
	end

	local _findfirst = alien.MSVCRT._findfirst
	local _findnext = alien.MSVCRT._findnext
	local _findclose = alien.MSVCRT._findclose

	_findfirst:types{ret="long",abi="cdecl", "string","pointer"}
	_findnext:types{ ret="int", abi="cdecl", "long",  "pointer"}
	_findclose:types{ret="int", abi="cdecl", "long"}

	hFile = _findfirst(path..pattern,_finddata_t());
	local tFiles = {}
	if hFile ~= -1 then
		tFiles[getfilename(_finddata_t)] = {
			["attrib"] = _finddata_t.attrib,
			["time_create"] = _finddata_t.time_create,
			["time_access"] = _finddata_t.time_access,
			["time_write"] = _finddata_t.time_write,
			["size"] = _finddata_t.size,
		}
		local iRes = _findnext(hFile,_finddata_t())
		while iRes == 0 do
			tFiles[getfilename(_finddata_t)] = {
				["attrib"] = _finddata_t.attrib,
				["time_create"] = _finddata_t.time_create,
				["time_access"] = _finddata_t.time_access,
				["time_write"] = _finddata_t.time_write,
				["size"] = _finddata_t.size,
			}
			iRes = _findnext(hFile,_finddata_t())
		end
		_findclose(hFile)
	end
	return tFiles 
end

_dcppmeta.Utils.formatBytes = function(i)
	i = tonumber(i) or 0
	argcheck("dcpp.Utils.formatBytes",1,i,{"number"})
	local j,tUnits = 1, {"B","Kb","Mb","Gb","Tb","Pb","Eb","Yb"}
	while i > 1024 do i = i/1024; j=j+1 end
	return string.format("%.2f",i).." "..(tUnits[j] or "??") 
end

_dcppmeta.Utils.formatClientType = function(ct)
	ct = tonumber(ct) or 0
	argcheck("dcpp.Utils.formatClientType",1,ct,{"number"})
	local t,i = {},ct
	if i >= 0x20 then table.insert(t,1,"Hub");        i=i-0x20; end
	if i >= 0x10 then table.insert(t,1,"Hubowner");   i=i-0x10; end
	if i >= 0x08 then table.insert(t,1,"Super User"); i=i-0x08; end
	if i >= 0x04 then table.insert(t,1,"Operator");   i=i-0x04; end
	if i >= 0x02 then table.insert(t,1,"Registered"); i=i-0x02; end
	if i >= 0x01 then table.insert(t,1,"Bot");        i=i-0x01; end
	local ret = (next(t) and table.concat(t," ") or "Normal user").." ("..(ct)..")"
	return ret:sub(1,1)..ret:sub(2):lower()
end

_dcppmeta.Utils.formatExactSize = function(i)
	i = tonumber(i) or 0
	argcheck("dcpp.Utils.formatExactSize",1,i,{"number"})
	local ret,num = "",tostring(i)
	while #num > 0 do
		ret = num:sub(-3,-1).." "..ret
		num = num:sub(1,-4)
	end
	return ret
end

_dcppmeta.Utils.formatParams = function(s,t,b)
	argcheck("dcpp.Utils.formatParams",1,s,{"string"})
	argcheck("dcpp.Utils.formatParams",2,t,{"table"})
	return (s:gsub("%%%[([^ ]-)%]",function(key) return t[key] or (b and "%["..key.."]" or "") end))
end

_dcppmeta.Utils.formatSeconds = function(aSec,supressHours)
	aSec = tonumber(aSec) or 0
	if (not supressHours) then
		return string.format("%02d:%02d:%02d",(aSec / (60*60)),((aSec / 60) % 60),(aSec % 60));
	else
		return string.format("%02d:%02d",(aSec / 60),(aSec % 60));
	end
end

_dcppmeta.Utils.formatStatus = function(iStatus)
	iStatus = tonumber(iStatus) or 0
	argcheck("dcpp.Utils.formatStatus",1,iStatus,{"number"})
	local status,i = {},iStatus
	if i >= 0x20 then table.insert(status,1,"NAT");      i=i-0x20; end
	if i >= 0x10 then table.insert(status,1,"TLS");      i=i-0x10; end
	if i >= 0x08 then table.insert(status,1,"Fireball"); i=i-0x08; end
	if i >= 0x04 then table.insert(status,1,"Server");   i=i-0x04; end
	if i >= 0x02 then table.insert(status,1,"Away");     i=i-0x02; end
	if i >= 0x01 then table.insert(status,1,"Normal");   i=i-0x01; end
	return (next(status) and table.concat(status," ") or "Unknown").." ("..(iStatus)..")"
end

_dcppmeta.Utils.getAppPath = function()
	return dco:GetAppPath()
end

_dcppmeta.Utils.getIpCountry = function(ip,fullname)
	argcheck("dcpp.Utils.getIpCountry",1,ip,{"string"})
	return libgeoip[fullname and "name" or "code"](ip)
end

_dcppmeta.Utils.getStartTime = function()
	return os.time() - _dcpp._start
end

_dcppmeta.Utils.getVersion = function()
	return tonumber(dcpp.SettingsManager.get("ConfigVersion") or 0.75)
end

_dcppmeta.Utils.ipToDec = function(ip)
	argcheck("dcpp.Utils.ipToDec",1,ip,{"string"})
	local a,b,c,d = dcpp.Utils.isIp(ip);
	return a and (a*0x1000000 + b*0x10000 + c*0x100 + d) or -1;
end

_dcppmeta.Utils.isIp = function(ip)
	argcheck("dcpp.Utils.isIp",1,ip,{"string"})
	local a,b,c,d = ip:match("^(%d+)%.(%d+)%.(%d+)%.(%d+)$");
	if a and tonumber(a) < 0xFF and tonumber(b) < 0xFF and tonumber(c) < 0xFF and tonumber(d) < 0xFF then
		return a,b,c,d;
	else return nil;
	end
end

_dcppmeta.Utils.isPrivateIp = function(ip)
	argcheck("dcpp.Utils.isPrivateIp",1,ip,{"string","number"})
	if type(ip) ~= "number" then
		ip = dcpp.Utils.ipToDec(ip)
	end
	if ip > -1 then
		if (ip > 0x0A000000 and ip < 0x0AFFFFFF) -- 10.0.0.0/8
		or (ip > 0xAC100000 and ip < 0xAC1FFFFF) -- 172.16.0.0/12
		or (ip > 0xC0A80000 and ip < 0xC0A8FFFF) -- 192.168.0.0/16
		or (ip > 0x7F000000 and ip < 0x7FFFFFFF) -- 127.0.0.0/8
		then return true end
	end
	return false
end

_dcppmeta.Utils.sendUDP = function(ip_port,msg)
	argcheck("dcpp.Utils.sendUDP",1,ip_port,{"string"})
	argcheck("dcpp.Utils.sendUDP",2,msg,{"string"})
	dco:SendUDP(ip_port,msg)
end

_dcppmeta.Utils.serialize = function(filename,tstore,tblname,tab)
	argcheck("dcpp.Utils.serialize",1,filename,{"string"})
	argcheck("dcpp.Utils.serialize",2,tstore,{"table","string"})
	argcheck("dcpp.Utils.serialize",3,tblname,{"string"})
	argcheck("dcpp.Utils.serialize",4,tab,{"string","nil"})
	tab = tab or "";
	if not _dcpp.__FILE__ then
		_dcpp._deep = 1
		_dcpp.__FILE__ = io.open(dcpp.Utils.validateFileName(filename),"w+")
	end
	_dcpp.__FILE__:write(tab..tblname.." = {\n")
	for key,value in dcpp.Utils.sortByKeys(tstore) do
		if (type(key) ~= "function" and type(value) ~= "function" and type(key) ~= "userdata" and type(value) ~= "userdata") then
			local s_key = ((type(key) == "number") and "[%d]" or "[%q]"):format(type(key) == "number" and key or tostring(key))
			if(type(value) == "table") then
				_dcpp._deep = _dcpp._deep + 1
				if _dcpp._deep < 15 then -- Avoid endless recursion (eg. _G._G._G._G.....) max step 15
					dcpp.Utils.serialize(filename,value,s_key,tab.."\t")
				end
			else
				local s_value = (type(value) == "string") and ("%q"):format(value) or tostring(value)
				_dcpp.__FILE__:write(tab.."\t"..s_key.." = "..s_value)
			end
			_dcpp.__FILE__:write(",\n")
		end
	end
	_dcpp._deep = _dcpp._deep - 1
	_dcpp.__FILE__:write(tab.."}")
	if tab == "" then
		_dcpp.__FILE__:close()
		_dcpp.__FILE__ = nil
		_dcpp._deep = nil
	end
end

_dcppmeta.Utils.sortByKeys = function(t,desc)
	local x,i = {},0
	for n in pairs(t) do table.insert(x,n) end
	local f = function(a,b)
		a = tonumber(a) or tostring(a) b = tonumber(b) or tostring(b)
		if type(a) ~= type(b) then a = tostring(a) b = tostring(b) end
		return (desc and a > b) or (not desc and a < b)
	end
	table.sort(x,f)
	return function() i = i+1; return x[i] and x[i],t[x[i]] or nil end
end

_dcppmeta.Utils.validateFileName = function(filename)
	argcheck("dcpp.Utils.validateFileName",1,filename,{"string"})
	local badChars = {["<"]=true,[">"]=true,["/"]=true,["\""]=true,["|"]=true,["?"]=true,["*"]=true}
	-- First, eliminate forbidden chars
	filename = filename:gsub("(.)",function(c) if c:byte() < 0x20 or badChars[c] then return "_" else return c end end)
	-- Then, eliminate all ':' that are not the second letter ("c:\...")
	filename = filename:sub(1,2)..filename:sub(3):gsub(":","_")
	-- Remove the .\ that doesn't serve any purpose
	filename = filename:gsub("\\%.\\",""):gsub("/%./","")
	-- Remove any double \\ that are not at the beginning of the path...
	filename = filename:sub(1,2)..filename:sub(3):gsub("\\\\",""):gsub("//","")
	-- And last, but not least, the infamous ..\! ...
	filename = filename:gsub("\\%.%.\\","___"):gsub("/%.%./","___")
	-- Dots at the end of path names aren't popular
	filename = filename:gsub("%.\\$","_"):gsub("%./$","_")
	return filename;
end

setmetatable(dcpp,{
__index = function(t,k)
	if k == "setListener" then
		print("You want to use an old script which is incompatible with this version. Please update it.")
	else
		return rawget(_dcppmeta,k)
	end
end,})

-- Create environment for scripts
local _ENVIRONMENT = {}

for var,val in pairs(_G) do -- create a shallow copy
	if var ~= "_ALERT" then
		rawset(_ENVIRONMENT,var,val)
	end
end

--[=[-------------]=]--
--[=[ Lua manager ]=]--
--[=[-------------]=]--

if not loadfile(dcpp.Utils.getAppPath().."scripts\\scriptlist.tbl") then
	local f = io.open(dcpp.Utils.getAppPath().."scripts\\scriptlist.tbl","w")
	if f then f:write("scriptlist = {\n}") f:close() end
end
dofile(dco:GetAppPath().."scripts\\scriptlist.tbl" )

-- localize
local tcopy = scriptlist
scriptlist = nil
local scriptlist = tcopy
tcopy = nil

local saveScripts = function()
	dcpp.Utils.serialize(dcpp.Utils.getAppPath().."scripts\\scriptlist.tbl",scriptlist,"scriptlist")
end

local removeListeners = function(filename)
	_dcpp._listeners[filename] = nil
	_ENV[filename] = nil -- Remove globals
end

local loadScript = function(filename)
	filename = filename:sub(1,-5)
	local ret = "has been successfully loaded."
   	if _dcpp._listeners[filename] then -- unload before load it again
		removeListeners(filename)
	  	ret = "has been successfully reloaded."
	else
		_dcpp._listeners[filename] = {} -- Load it, even if it doesn't have listeners to be able to unload later
	end
    local fnc, err = loadfile(dcpp.Utils.getAppPath().."scripts\\"..filename..".lua")
    if not fnc then
		return nil,"Unable to load "..filename..".lua because: "..err
    end
	scriptlist[filename] = 1
	saveScripts()
	_ENV[filename] = {}
	_ENV[filename]["OnUserCommand"] = function(...)
		local tArgs = {...}
		tArgs[1] = dcpp.ClientManager.findHub(tArgs[1])
		_ENV[filename]["OnOwnMessage"](tArgs[1],unpack(tArgs,2))
	end
	for a,b in pairs(_ENVIRONMENT) do _ENV[filename][a] = b end -- Global env's shallow copy
	_ENV[filename]["_G"] = _ENV[filename] -- Script's global variables' storage
	_ENV[filename]["dofile"] = function(file) -- Push chunk to own ENV
		local fnc,err = loadfile(file)
		if fnc then debug.setfenv(fnc,_ENV[filename])()
		else error(err)
		end
	end
	_ENV[filename]["error"] = _ALERT -- Own error handler
	debug.setfenv(fnc,_ENV[filename])()
    local listeners = {
		["OnConnected"] = true,
		["OnDisconnected"] = true,
		["OnUserConnected"] = true,
		["OnUserDisconnected"] = true,
		["OnChatMessage"] = true,
		["OnOwnMessage"] = true,
		["OnPrivateMessageIn"] = true,
		-- TODO: ["OnPrivateMessageOut"] = true,
		["OnInfo"] = true,
		["OnHubRaw"] = true,
		["OnRawIn"] = true,
		["OnRawOut"] = true,
		["OnTimer"] = true,
	}
	for event in pairs(listeners) do
		if type(_ENV[filename][event]) == "function" then
			if not _dcpp._listeners[filename] then _dcpp._listeners[filename] = {} end
			_dcpp._listeners[filename][event] = _ENV[filename][event] -- Store event in private table
		end
	end
	collectgarbage("collect")
	return ret,err
end

local unloadScript = function(filename)
	filename = filename:sub(1,-5)
	if not _dcpp._listeners[filename] then return "not found." end
	local mem = collectgarbage("count")
	removeListeners(filename)
	-- change status flag in list
	scriptlist[filename] = 0
	-- remove usercommands
	for id,uc in pairs(_dcpp._usercommands) do
		if uc[2] == filename then _dcpp._usercommands[id] = nil end
	end
	saveScripts()
	collectgarbage("collect")
	return "unloaded, "..dcpp.Utils.formatBytes((mem-collectgarbage("count"))*1024).." of memory freed"
end

local removeScript = function(filename)
	unloadScript(filename)
	filename = filename:sub(1,-5)
	if scriptlist[filename] then
		scriptlist[filename] = nil
		saveScripts()
		return "successfully removed."
	end
	return "not found."
end

local OnEvent = function(functionname, ... )
	local ret = nil
	for script,events in pairs(_dcpp._listeners) do
		if events[functionname] then
			ret = events[functionname](...) or ret
		end
	end
	return ret
end

-- Load scriptmanager
_dcpp._listeners["ScriptManager"] = {}
_dcpp._listeners["ScriptManager"]["OnOwnMessage"] = function(hub, message)
	if message:sub(1,1) ~= "/" then return nil end
	local params = dcpp.Text.tokenize(message)
	if params[1] == "/lualist" then
		local t,msg,i = {},"",0
		for a in pairs(dcpp.Utils.forFiles(dcpp.Utils.getAppPath().."scripts","*.lua",0)) do
			if scriptlist[a:sub(1,-5)] then i=i+1 end
			t[a] = ("- [ %d ] %s"):format(scriptlist[a:sub(1,-5)] and 1 or 0,a)
		end
		for a,b in dcpp.Utils.sortByKeys(t) do msg = msg..b.."\r\n" end
		hub:addHubLine(i.." scripts loaded:\r\n"..msg )
		return 1
	elseif params[1] == "/luaload" then
		if params[2]:sub(-4,-1) ~= ".lua" then
			hub:addHubLine("You can load just regular Lua scripts")
			return 1
		end
		if params[2] == "startup.lua" then
			hub:addHubLine("If you want to reload startup.lua, restart your client (hence most script fail)")
			return 1
		end
		local ret, err = loadScript(params[2])
		hub:addHubLine((ret and "Lua script '"..params[2].."' "..ret) or err )
		return 1
	elseif params[1] == "/luaunload" then
		if params[2]:sub(-4,-1) ~= ".lua" then
			hub:addHubLine("You can unload just regular Lua scripts")
			return 1
		end
		if params[2] == "formatting.lua" then
			hub:addHubLine("You can't unload formatting.lua because you can't see the chat messages without that.")
			return 1
		elseif params[2] == "startup.lua" then
			hub:addHubLine("Why? bzzz...")
			return 1
		end
		local ret = unloadScript( dcpp.Text.utf8ToAcp(dcpp.Text.acpToUtf8(params[2])) )
		hub:addHubLine( "Lua script '"..params[2].."' "..ret )
		return 1
	elseif params[1] == "/luaremove" then
		if params[2]:sub(-4,-1) ~= ".lua" then
			hub:addHubLine("You can load just regular Lua scripts")
			return 1
		end
		if params[2] == "formatting.lua" then
			hub:addHubLine("You can't remove formatting.lua because you can't see the chat messages without that.")
			return 1
		elseif params[2] == "startup.lua" then
			hub:addHubLine("If you want to reload startup.lua, restart your client (hence all your scripts fails)")
			return 1
		end
		local ret = removeScript(params[2])
		hub:addHubLine( "Lua script '"..params[2].."' "..ret )
		return 1
	elseif params[1] == "/meminfo" then
		hub:addHubLine( "Scripts are using "..dcpp.Utils.formatBytes(collectgarbage("count")*1024).." of memory." )
		return 1
	elseif params[1] == "/help" then
		hub:addHubLine("<startup.lua> /lualist, /luaload <filename>, /luaunload <filename>, /luaremove <filename>, /meminfo" )
		return nil
	end
end

--[=[----------------------------------------]=]--
--[=[ User instance & $MyINFO || BINF parser ]=]--
--[=[----------------------------------------]=]--

local updateFields = function(user,inf)
	if user:isAdc() then
		user:set("SID",inf:sub(6,9))
		for s in inf:sub(10):gmatch("([^ ]+)") do
			user:set(s:sub(1,2),s:sub(3))
		end
		if user:get("SU"):match("UDP") then user:set("MO","A") else user:set("MO","P") end
		if (user:isSet("VE") and user:isSet("HN") and user:isSet("HR") and user:isSet("HO") and user:isSet("SL")) then
			user:set("TA","<"..user:get("VE")..",M:"..user:get("MO")..",H:"..
			user:get("HN").."/"..user:get("HR").."/"..user:get("HO")..",S:"..user:get("SL")..">")
		end
		if user:isSet("CT") and tonumber(user:get("CT")) > 3 then user:set("OP","1") end
	else
		-- NMDC2ADC
		local NI,DE,CO,CT,EM,SS = inf:match("^%$MyINFO %$ALL ([^ ]+) ([^%$]*)%$ %$([^%$]*)(.)%$([^%$]*)%$([^%$]*)%$")
		if CT and CT:byte() > 63 then CO = CO..CT end
		local TA = ""
		if DE then
			TA = DE:match("(%b<>)$") or ""
			DE = DE:gsub("%b<>$","")
		end
		if SS == "" then SS = "0" end
		if CT then CT = tostring(string.byte(CT)) end
		local VE,MO,HN,HR,HO,SL,US
		if TA ~= "" then
			VE = TA:match("^<(.- V:[^,]+),") or ""
			MO = TA:match(",M:([^,]+),") or ""
			HN = TA:match(",H:(%P)") or ""
			HR = TA:match(",H:%P%p(%P)%p%P,") or ""
			HO = TA:match(",H:%P%p%P%p(%P),") or ""
			SL = TA:match(",S:([^,>]+)") or ""
			US = TA:match(",[BL]:(%d+)") or ""
			if US ~= "" then US = tonumber(US)*1024 end
		end
		user:set("NI",NI); user:set("DE",DE); user:set("TA",TA); user:set("CO",CO);
		user:set("CT",CT); user:set("EM",EM); user:set("SS",SS); user:set("VE",VE);
		user:set("MO",MO); user:set("HN",HN); user:set("HR",HR); user:set("HO",HO);
		user:set("SL",SL); user:set("US",US);
	end
end

local createUser = function(hubobj,hub,id,inf)
	
	local user = {}

	user._hub = hubobj
	user._fields = {["LT"] = os.time()}

	local get = function(self,field)
		argcheck("user:get",1,field,{"string"})
		return user._fields[field] or ""
	end
	
	local getCID = function(self)
		return self:get("ID")
	end

	local getClient = function(self)
		return user._hub
	end

	local getParams = function(self,prefix)
		argcheck("user:getParams",1,prefix,{"string"})
		local t = {}
		for a,b in pairs(user._fields) do
			t[prefix..a] = b
		end
		return t
	end

	local getSIDString = function(self)
		return user._fields["SID"] or ""
	end

	local getStatus = function(self)
		return self:get("CT")
	end

	local getTag = function(self)
		return self:get("TA")
	end

	local isAdc = function(self)
		return user._hub:isAdc()
	end

	local isClientType = function(self,status)
		argcheck("user:isClientType",1,status,{"number"})
		local ct = self:get("CT")
		return tonumber(ct) == status
	end

	local isSet = function(self,field)
		argcheck("user:isSet",1,field,{"string"})
		return self:get(field) ~= ""
	end
	
	local set = function(self,field,value)
		argcheck("user:set",1,field,{"string"})
		value = self:getClient():escape(tostring(value),true)
		user._fields[field] = value
	end
	
	
	local userinstance = {
		get = get,             getClient = getClient, getParams = getParams, getSIDString = getSIDString, 
		getStatus = getStatus, getTag = getTag,       isAdc = isAdc,         isClientType = isClientType, 
		isSet = isSet,         set = set,}
		
	local mt = {
		__newindex = function()
			print("Attempt to write into the protected user table!")
		end,
		__metatable = "Mind your own business",
		}
	setmetatable(userinstance,mt)
	hub._users[id] = userinstance
	if type(inf) == "string" then
		updateFields(hub._users[id],inf)
	end
	return userinstance
end

--[=[--------------]=]--
--[=[ Hub instance ]=]--
--[=[--------------]=]--

local createHub = function(hubid)
	local hub = {}
	hub._myID = ""
	hub._connected = false
	hub._users = {}
	hub._uptime = os.time()
	hub._fields = {["URL"] = dco:GetHubIpPort(hubid)}

	local addHubLine = function(self,msg)
		argcheck("hub:addHubLine",1,msg,{"string"})
		if self:isAdc() then
			dco:InjectHubMessage(hubid,"* "..msg) -- w/o the star, the client crashes :-/
		else
			dco:InjectHubMessage(hubid,msg)
		end
	end

	local attention = function(self)
		--dco:HubWindowAttention(hubid)
	end

	local escape = function(self,msg,inverse)
		argcheck("hub:escape",1,msg,{"string"})
		if self:isAdc() then
			msg = msg:gsub("\r","")
			if inverse then
				return msg:gsub([[\\]],[[\]]):gsub([[\s]]," "):gsub([[\n]],"\n")
			else
				return msg:gsub([[\]],[[\\]]):gsub(" ",[[\s]]):gsub("\n",[[\n]])
			end
		else
			if inverse then
				return msg:gsub("&#%d+;",function(s) return ("%c"):format(tonumber(s)) end)
			else
				return msg:gsub("[%$|]",function(s) return ("&#%i;"):format(s:byte()) end)
			end
		end
	end

	local findUser = function(self,p1,p2)
		-- Inner use
		if not hub._users[p1] and p2 and type(p2) ~= "boolean" then
			hub._users[p1] = createUser(self,hub,p1,p2) -- Create user from INFO
			if self:getUptime() >= 2 then -- start sending messages AFTER logon
				OnEvent("OnUserConnected",self,hub._users[p1])
			end
			return hub._users[p1]
		else
			if p2 == true then -- Force create a new user from scratch
				createUser(self,hub,p1):set("NI",p1) -- on ADC it's not right, but BINF will change
				if hub._myID == "" then hub._myID = p1 end
			elseif p2 == false then -- Remove the user
				hub._users[p1] = nil
			elseif p2 then -- INFO (not first)
				if p2:sub(1,7) ~= "$OpList" then
					updateFields(hub._users[p1],p2)
				end
				if not hub._connected and ((self:isAdc() and p1 == hub._myID) or (not self:isAdc() and p2:sub(1,7) == "$OpList")) then 
					hub._connected = true -- We got our INFO, we are connected
					if _dcpp._oplists[self:getHubUrl()] then
						for nick in msg:sub(9):gmatch("[^$]+") do
							local u = h:findUser(nick,msg)
							if u then u:set("OP","1") end
						end
						_dcpp._oplists[self:getHubUrl()] = nil
					end
					for i,uc in dcpp.Utils.sortByKeys(_dcpp._usercommands) do
						if not uc[1]:isHub() then
							local address = uc[1]:getHub()
							local url = self:getAddress()
							local adc = self:isAdc()
							if address == "" then address = url
							elseif address == "op" then
								if ((adc and p2:match(" OP1")) or (not adc and p2:find(hub._myID.."$$",8,1))) then
									address = url
								end
							end
							if address == url then
								local cmd = uc[1]:getCommand()
								cmd = cmd:gsub("(OnOwnMessage%()(\"\"),","%1dcpp.ClientManager.findHub(\""..address.."\"),")
								if adc then
									self:parseCommand("ICMD "..self:escape(uc[1]:getName()).." TT"..
									self:escape(cmd).." CT"..uc[1]:getContext())
								else
									self:parseCommand("$UserCommand "..uc[1]:getType().." "..uc[1]:getContext()..
									" "..self:escape(uc[1]:getName()).."$"..self:escape(cmd))
								end
							end
						end
					end
				end
			end
		end
		-- No 2nd parameter: get user
		argcheck("hub:findUser",1,p1,{"string","number"})
		if self:isAdc() then
			if type(p1) == "number" then
				local sid = dcpp.AdcCommand.fromSID(p1)
				return hub._users[sid]
			end
			for id,u in pairs(hub._users) do -- Maybe we want a user by nick (but id is SID there)
				if u:get("NI"):lower() == p1:lower() then return u end
			end
			if p1:match("^"..("[A-Z2-7]"):rep(39).."$") then -- by CID
				for id,u in pairs(hub._users) do
					if u:get("ID") == p1 then return u end
				end
			end
		end
		return hub._users[p1]
	end

	local getAddress = function(self)
		return dco:GetHubIpPort(hubid)
	end

	local getAvailable = function(self)
		local i = 0
		for _,u in pairs(hub._users) do i = i + (tonumber(u:get("SS")) or 0) end
		return i
	end

	local getHubName = function(self)
		return self:getHubIdentity():get("NI")
	end

	local getHubDescription = function(self)
		return self:getHubIdentity():get("DE")
	end

	local getHubIdentity = function(self)
		return {
		get = function(this,param)
			argcheck("hubIdentity:get",1,param,{"string"})
			return hub._fields[param] or ""
		end,
		getCID = function(this)
			return "" -- hub doesn't have CID
		end,
		getClient = function(this)
			return self
		end,
		getParams = function(this,prefix)
			argcheck("hubIdentity:getParams",1,prefix,{"string"})
			local t = {}
			for a,b in pairs(hub._fields) do
				t[prefix..a] = b
			end
			return t
		end,
		getSIDString = function(this)
			return "" -- hub doesn't have SID
		end,
		getStatus = function(this)
			return this:get("CT")
		end,
		getTag = function(this)
			return this:get("TA")
		end,
		isAdc = function(this)
			return self:isAdc()
		end,
		isClientType = function(this,status)
			argcheck("hubIdentity:isClientType",1,status,{"number"})
			local ct = this:get("CT")
			return tonumber(ct) == status
		end,
		isSet = function(this,param)
			argcheck("hubIdentity:isSet",1,param,{"string"})
			return this:get(param) ~= ""
		end,
		set = function(this,param,value)
			argcheck("hubIdentity:set",1,param,{"string"})
			argcheck("hubIdentity:set",2,value,{"string","number"})
			value = self:escape(tostring(value),true)
			hub._fields[param] = value
		end,}
	end

	local getHubUrl = function(self)
		return dco:GetHubUrl(hubid)
	end

	local getIp = function(self)
		return self:getAddress():match("([^ ])[:]?%d*")
	end

	local getMyIdentity = function(self)
		return hub._users[hub._myID]
	end

	local getPort = function(self)
		local port = self:getAddress():match("[^ ][:]?(%d*)")
		return port ~= "" and tonumber(port) or 411
	end

	local getUptime = function(self)
		return os.time() - hub._uptime
	end

	local getUserCount = function(self)
		local i = 0
		for _ in pairs(hub._users) do i = i+1 end
		return i
	end

	local getUserList = function(self)
		local t = {}
		for a,b in pairs(hub._users) do
			t[a] = b
		end 
		return t
	end

	local hubMessage = function(self,msg,tp)
		argcheck("hub:hubMessage",1,msg,{"string"})
		if self:isAdc() then
			local ownSid = self:getMyIdentity():get("SID")
			dco:SendHubMessage(hubid, string.format("BMSG %s %s %s\n",ownSid,self:escape(msg),tp and "ME1" or "\008"))
		else
			msg = dcpp.Text.utf8ToAcp(msg)
			local nick = self:getMyIdentity():get("NI")
			dco:SendHubMessage(hubid, string.format("<%s> %s%s|",nick,tp and "/me" or "",self:escape(msg)))
		end		
	end

	local isAdc = function(self)
		local url = getHubUrl()
		return (url and (url:sub(1,6) == "adc://" or url:sub(1,7) == "adcs://"))
	end

	local isConnected = function(self,id)
		return hub._connected
	end

	local isOp = function(self)
		return self:getMyIdentity():get("OP") == 1
	end

	local parseCommand = function(self,msg)
		argcheck("hub:parseCommand",1,msg,{"string"})
		if self:isAdc() then
			dco:InjectHubMessageADC(hubid,msg)
		else
			dco:InjectHubMessage(hubid,msg)
		end
	end

	local privateMessage = function(self,ou,msg,tp,hfs)
		argcheck("hub:privateMessage",1,ou,{"table"})
		argcheck("hub:privateMessage",2,msg,{"string"})
		if self:isAdc() then
			local ownSid = self:getMyIdentity():get("SID")
			local victimSid = ou:get("SID")
			if ownSid then
				dco:SendHubMessage(hubid,("%sMSG %s %s %s %s PM%s\n"):format(hfs and "D" or "E",ownSid,victimSid,self:escape(msg),tp and "ME1" or "\008",ownSid))
			end	
		else
			local ownNick = self:getMyIdentity():get("NI")
			if ownNick then
				local victim = ou:get("NI")
				dco:SendHubMessage(hubid,("$To: %s From: %s $<%s> %s%s|"):format(victim,ownNick,ownNick,tp and "/me" or "",self:escape(msg)))
				if not hfs then
					self:parseCommand(("$To: %s From: %s $<%s> %s%s"):format(victim,ownNick,ownNick,tp and "/me" or "",self:escape(msg)))
				end
			end	
		end
	end

	local redirect = function(self,address)
		argcheck("hub:redirect",1,address,{"string"})
		if self:isAdc() then
			self:parseCommand("IQUI "..self:getMyIdentity():get("SID").." RD"..address)
		else
			self:parseCommand("$ForceMove "..address)
		end
	end
	
	local sendUserCmd = function(self,cmd)
		argcheck("hub:sendUserCmd",1,cmd,{"string"})
		if self:isAdc() then
			cmd = dcpp.Text.acpToUtf8(cmd)
			if cmd:sub(-1,-1) ~= "\n" then cmd = cmd.."\n" end
		else
			if cmd:sub(-1,-1) ~= "|" then cmd = cmd.."|" end
		end
		dco:SendHubMessage(hubid,cmd)
	end

	local hubinstance = {
		addHubLine = addHubLine,         attention = attention,                 escape = escape,
		findUser = findUser,             getAddress = getAddress,               getAvailable = getAvailable,
		getHubName = getHubName,         getHubDescription = getHubDescription, getHubIdentity = getHubIdentity,
		getHubUrl = getHubUrl,           getIp = getIp,                         getMyIdentity = getMyIdentity,
		getPort = getPort,               getUptime = getUptime,                 getUserCount=getUserCount,
		getUserList = getUserList,       hubMessage = hubMessage,               isAdc = isAdc,
		isConnected = isConnected,       isOp = isOp,                           parseCommand = parseCommand,
		privateMessage = privateMessage, redirect = redirect,                   sendUserCmd = sendUserCmd,}
		
	local mt = {
		__newindex = function()
			print("Attempt to write into the protected hub table!")
		end,
		__metatable = "Mind your own business",
		}
	setmetatable(hubinstance,mt)
	return hubinstance
end

--[=[-------------------------]=]--
--[=[ UserConnection instance ]=]--
--[=[-------------------------]=]--

local createUserConnection = function(user)
	local close = function(self)
		dco:DropUserConnection(user)
	end
	local getIp = function(self)
		return dco:GetClientIp(user)
	end
	local send = function(self,message)
		argcheck("userconnection:send",1,message,{"string"})
		if message:sub(-1,-1) ~= "|" then message = message.."|" end
		dco:SendClientMessage(user,message)
	end
	local userconnectioninstance = {close = close, getIp = getIp, send = send}
	local mt = {
		__newindex = function()
			print("Attempt to write into the protected userconnection table!")
		end,
		__metatable = "Mind your own business",
		}
	setmetatable(userconnectioninstance,mt)
	return userconnectioninstance
end

--[=[----------------]=]--
--[=[ Event handlers ]=]--
--[=[----------------]=]--

nmdch = {}

function nmdch.DataArrival(hub,msg)
	local h = dcpp.ClientManager.findHub(hub)
	if not h then return false end -- Never should happen
	if OnEvent("OnHubRaw",h,msg) then
		return true
	end
	local cmd = msg:match("^%$([^ ]+)")
	if cmd then
		if cmd == "Hello" then
			local nick = msg:sub(8)
			h:findUser(nick,true) -- Create user functions, fields are empty (no MyINFO so far)
		elseif cmd == "HubName" then
			local name,desc = msg:sub(10):match("(.-) %- (.+)")
			if name then
				h:getHubIdentity():set("NI",name)
				h:getHubIdentity():set("DE",desc)
			else
				h:getHubIdentity():set("NI",msg:sub(10))
			end
			return nil
		elseif cmd == "MyINFO" and msg:sub(1,13) == "$MyINFO $ALL " then
			local nick = msg:match("^%$MyINFO %$ALL ([^ ]+) ")
			if nick then
				local user = h:findUser(nick,msg) -- create or update user
				return OnEvent("OnInfo",h,user,msg)
			end
		elseif cmd == "OpList" then
			if not h:isConnected() then
				for nick in msg:sub(9):gmatch("[^$]+") do
					local u = h:findUser(nick,msg)
					if u then u:set("OP","1") end
				end
			else
				_dcpp._oplists[h:getHubUrl()] = msg
			end
			return nil
		elseif cmd == "Quit" then
			local nick = msg:sub(7)
			local u = h:findUser(nick)
			if not u then u = h:findUser(nick,true) end
			OnEvent("OnUserDisconnected",h,u)
			h:findUser(nick,false) -- delete just after the scripts parsed
			return nil
		elseif cmd == "To:" then
			local to,from,replyto,text = msg:match("^%$To: ([^ ]+) From: ([^ ]+) %$<([^ ]+)> (.*)$")
			if to then
				local uto = h:findUser(to); if not uto then uto = h:findUser(to,true) end
				local ufrom = h:findUser(from); if not ufrom then ufrom = h:findUser(from,true) end
				local urlt = h:findUser(replyto); if not urlt then utrlt = h:findUser(replyto,true) end
				return OnEvent("OnPrivateMessageIn",h,ufrom,uto,urlt,h:escape(text,true),text:sub(1,3)=="/me")
			end
		elseif cmd == "UserCommand" then
			local UCType,UCContext = msg:match("^%$UserCommand (%d+) (%d+)")
			local name,command = msg:match("^%$UserCommand %d+ %d+ ([^%$]-)%$(.+)$")
			if UCType and name then
				dcpp.FavoriteManager.addUserCommand(tonumber(UCType),tonumber(UCContext),name,command,h,true)
			end
			return true
		elseif cmd == "UserIP" then
			for nick_ip in msg:sub(9):gmatch("[^$]+") do
				local nick,ip = nick_ip:match("^([^ ]+) ([^ ]+)$" )
				if nick then
					local u = h:findUser(nick)
					if u then u:set("I4",ip) 
					else u = h:findUser(nick,true) u:set("I4",ip)
					end
				end
			end
			return nil
		end
	elseif msg:sub(1,1) == "<" then
		local nick,text = msg:match("^<([^ ]+)> (.*)$" )
		if nick then
			local u = h:findUser(nick)
			if not u then -- doesn't have table -> user w/o MyINFO, create empty user from nick
				u = h:findUser(nick,true)
			end
			return OnEvent("OnChatMessage",h,u,h:escape(text,true),false)
		end
	elseif msg:sub(1,1) == "*" then
		local nick,text = msg:match("^* ([^ ]+) (.*)")
		if nick then
			local u = h:findUser(nick)
			if not u then u = h:findUser(nick,true) end
			return OnEvent("OnChatMessage",h,u,h:escape(text,true),true)
		end
	end
end

function nmdch.OnHubAdded(hub)
	_dcpp._hubs[hub] = createHub(hub)
	OnEvent("OnConnected",_dcpp._hubs[hub])
end

function nmdch.OnHubRemoved(hub)
	if _dcpp._hubs[hub] then
		OnEvent("OnDisconnected",_dcpp._hubs[hub])
		_dcpp._hubs[hub] = nil
	end
end

adch = {}

function adch.DataArrival(hub,msg)
	if msg == "" then
		return nil
	end
	--msg = dcpp.Text.utf8ToAcp(msg) -- It'll fix the locale-supported characters (no utf8 in Lua by default)
	local h = _dcpp._hubs[hub]
	if OnEvent("OnHubRaw",h,msg) then
		return 1
	end
	local params = dcpp.Text.tokenize(msg)
	local mtype = params[1]:sub(1,1)
	local cmd = params[1]:sub(2,4)
	local sid, targetSid, parameters = nil, nil, ""
	if mtype == "B" then
		parameters = msg:sub(11)
		sid = params[2]
	elseif mtype == "D" or mtype == "E" then
		parameters = msg:sub(16)
		sid = params[2]
		targetSid = params[3]
	end
	
	if cmd == "CMD" and mtype == "I" then
		local UCType,UCContext = 0,tonumber(msg:match(" CT(%d+)"))
		local name,command = msg:match("^ICMD ([^ ]+)"),msg:match(" TT([^ ]+)")
		dcpp.FavoriteManager.addUserCommand(UCType,UCContext,name,command,h,true)
		return true
	elseif cmd == "INF" then
		if mtype == "I" then -- INF is about the hub itself
			msg = msg:sub(5)
			for s in msg:gmatch("([^ ]+)") do
				h:getHubIdentity():set(s:sub(1,2),s:sub(3))
			end
			return nil
		else
			local u = h:findUser(params[2],msg) -- create or update user
			return OnEvent("OnInfo",h,u,msg)
		end
	elseif cmd == "MSG" then
		local istp,pm = false,false
		if parameters:match(" ME1") then istp = true end
		if parameters:match(" PM") then pm = true end
		if pm then
			local from,to,replyto,text = params[2],params[3],parameters:match(" PM(%w%w%w%w)"),params[4]
			local uto = h:findUser(dcpp.AdcCommand.toSID(to)); if not uto then uto = h:findUser(to,true) end
			local ufrom = h:findUser(dcpp.AdcCommand.toSID(from)); if not ufrom then ufrom = h:findUser(from,true) end
			local urlt = h:findUser(dcpp.AdcCommand.toSID(replyto)); if not urlt then utrlt = h:findUser(replyto,true) end
			return OnEvent("OnPrivateMessageIn",h,ufrom,uto,urlt,h:escape(text,true),istp)
		else
			if mtype == "B" then
				local sid,text = params[2],params[3]
				local u = h:findUser(dcpp.AdcCommand.toSID(sid))
				if not u then h:findUser(sid,true) end
				return OnEvent("OnChatMessage",h,u,h:escape(text,true),istp)
			elseif mtype == "I" then
				local text = params[2]
				local u = h:getHubIdentity()
				if not u then return false end
				return OnEvent("OnChatMessage",h,u,h:escape(text,true),istp)
			end
		end
	elseif cmd == "QUI" then
		local sid = params[2]
		local u = h:findUser(sid)
		if not u then u = h:findUser(sid,true) end
		OnEvent("OnUserDisconnected",h,u)
		h:findUser(sid,false) -- delete just after the scripts parsed
	elseif cmd == "SID" and mtype == "I" then
		h:findUser(msg:sub(6),true) -- Create user functions, fields are empty (no BINF so far)
	end
end

function adch.OnHubAdded(hub)
	_dcpp._hubs[hub] = createHub(hub)
	OnEvent("OnConnected",_dcpp._hubs[hub])
end

function adch.OnHubRemoved(hub)
	if _dcpp._hubs[hub] then -- In case the connection failed
		OnEvent("OnDisconnected",_dcpp._hubs[hub])
		_dcpp._hubs[hub] = nil
	end
end

function dcpp.UserDataIn(user,msg)
	local u = createUserConnection(user)
	return OnEvent("OnRawIn",u,msg)
end

function dcpp.UserDataOut(user,msg)
	local u = createUserConnection(user)
	return OnEvent("OnRawOut",u,msg)
end

function dcpp.OnCommandEnter(hub,text)
	return OnEvent("OnOwnMessage",_dcpp._hubs[hub],text)
end

function dcpp.OnTimer()
	OnEvent("OnTimer",1)
end


setmetatable(dcpp,{
	__index = function(t,k)
		if k == "setListener" then
			print("You want to use an old script which is incompatible with this version. Please update it.")
		else
			return rawget(_dcppmeta,k)
		end
	end,
	__newindex = function(t,k,v)
		if k == "FormatChatText" then -- only formatting.lua has acces to it
			rawset(t,k,v)
		else
			print("You can't store values in the main dcpp table!")
		end
	end,
	__metatable = "Mind your own business",
})

-- Always run timer
dco:RunTimer(1)

collectgarbage("collect")
local ret,err = loadScript("formatting.lua")
if not ret then error(err) end
-- autostart last running scripts
for script,loaded in pairs(scriptlist) do
	if loaded == 1 and script ~= "formatting" then
		local ret,err = loadScript(script..".lua")
		if ret then
			print(" ** Lua script '"..script.."' successfully loaded.")
		else
			print(err)
		end
	end
end
